import os
import numpy as np
from ase import Atoms
from ase.data import covalent_radii
from aegon.libutils import align
#------------------------------------------------------------------------------------------
def make_cell_vectors(atoms, latsp):
    positions = atoms.get_positions()
    min_coords = np.min(positions, axis=0)
    max_coords = np.max(positions, axis=0)
    cell_vectors = np.diag(max_coords - min_coords + latsp)
    return cell_vectors
#------------------------------------------------------------------------------------------
def tag(poscarlist):
    if not isinstance(poscarlist, list): poscarlist = [poscarlist]
    moleculeout = []
    for atoms in poscarlist:
        force = hasattr(atoms, "forces")
        numbers = list(set(atoms.get_atomic_numbers()))
        tagdict = {zi: itag for itag, zi in enumerate(sorted(numbers))}
        enumerate_molecule = atoms.copy()
        for iatom in enumerate_molecule:
            iatom.tag = tagdict[iatom.number]
        if force:
            enumerate_molecule.arrays['forces'] = atoms.arrays['forces']
        moleculeout.append(enumerate_molecule)
    return moleculeout
#------------------------------------------------------------------------------------------
def order_and_tag(poscarlist):
    if not isinstance(poscarlist, list): poscarlist = [poscarlist]
    moleculeout = []
    for atoms in poscarlist:
        numbers = list(set(atoms.get_atomic_numbers()))
        tagdict = {zi: itag for itag, zi in enumerate(sorted(numbers))}
        lista=[]
        force = hasattr(atoms, "forces")
        if force:
            allforces=atoms.arrays['forces']
            for k, katom in enumerate(atoms):
                n  = katom.number
                lista.append([tagdict[n], katom, allforces[k]])
        else: 
            for k, katom in enumerate(atoms):
                n  = katom.number
                lista.append([tagdict[n], katom]) 
        sorted_lista = sorted(lista, key = lambda x: x[0])
        sorted_forces=[]
        sorted_molecule=atoms.copy()
        for i, iatom in enumerate(sorted_molecule):
            iatom.tag=sorted_lista[i][0]
            xatom    =sorted_lista[i][1]
            iatom.symbol  =xatom.symbol
            iatom.position=xatom.position
            if force:
                sorted_forces.append(sorted_lista[i][2])
        if force:
            sorted_molecule.arrays['forces']=sorted_forces
        moleculeout.append(sorted_molecule)
    return moleculeout
#------------------------------------------------------------------------------------------
def molecule2poscar(atomslist, latsp=5.0):
    if not isinstance(atomslist, list): atomslist = [atomslist]
    moleculeout=[]
    for iatoms in atomslist:
        force = hasattr(iatoms, "forces")
        singleposcar=iatoms.copy()
        align(singleposcar)
        cell_vectors=make_cell_vectors(singleposcar, latsp)
        singleposcar.set_cell(cell_vectors)
        #vc=np.diag(cell_vectors)/2.0
        #singleposcar.translate(+vc)
        singleposcar.center()
        singleposcar.set_pbc(True)
        if force:
            singleposcar.arrays['forces'] = iatoms.arrays['forces']
        moleculeout.extend([singleposcar])
    moleculeout=order_and_tag(moleculeout)
    return moleculeout
#------------------------------------------------------------------------------------------
def readposcars(filename):
    if not os.path.isfile(filename):
        print("The file %s does not exist." %(filename))
        exit()
    contcarfile=open(filename,'r')
    poscarout=[]
    for line in contcarfile:
        header=line.split()
        name=header[0]
        energy=float(header[1]) if len(header)>1 else 0.0
        line=contcarfile.readline()
        if str(line.split()[0])=='0.00000000E+00': break
        #-----------------------------------
        scale=float(line.split()[0])
        line=contcarfile.readline()
        a1x, a1y, a1z=map(float,line.split())
        line=contcarfile.readline()
        a2x, a2y, a2z=map(float,line.split())
        line=contcarfile.readline()
        a3x, a3y, a3z=map(float,line.split())
        lattice_vectors=np.array([[a1x, a1y, a1z],[a2x, a2y, a2z],[a3x, a3y, a3z]])*scale
        #-----------------------------------
        line=contcarfile.readline()
        elements=line.split()
        line=contcarfile.readline()
        ocupnumchar=line.split()
        ocupnuminte=list(map(int, ocupnumchar))
        #-----------------------------------
        natom=sum(ocupnuminte)
        liste,kk=[],0
        for ii in ocupnuminte:
            for jj in range(ii):
                liste.append(elements[kk])
            kk=kk+1
        #-----------------------------------
        line=contcarfile.readline()
        sd=0
        if 'Selective dynamics' in line:
            line=contcarfile.readline()
            sd=1
        symbols=[]
        positions=[]
        if 'Direct' in line:
            for iatom in range(natom):
                line=contcarfile.readline()
                vecxyz=line.split()
                s=liste[iatom]
                symbols.append(s)
                xd=float(vecxyz[0])
                yd=float(vecxyz[1])
                zd=float(vecxyz[2])
                direct_coords = np.array([xd, yd, zd])
                cart_coords = np.dot(direct_coords, lattice_vectors)
                positions.append(cart_coords)
        if 'Cartesian' in line:
            for iatom in range(natom):
                s=liste[iatom]
                symbols.append(s)
                line=contcarfile.readline()
                cart_coords =np.array(line.split()[0:3])
                positions.append(cart_coords)
        positions=np.array(positions)
        iposcar = Atoms(symbols=symbols, positions=positions, cell=lattice_vectors, pbc=True)
        iposcar.info['i']=name
        iposcar.info['e']=energy
        poscarout.extend([iposcar])
    contcarfile.close()
    return poscarout
#------------------------------------------------------------------------------------------
def writeposcars(poscarlist, file, opt='D'):
    f=open(file,"w")
    for atoms in poscarlist:
        print("%s %12.8f" %(atoms.info['i'], atoms.info['e']), file=f)
        print('1.0', file=f)
        matrix=atoms.cell
        print("%20.16f %20.16f %20.16f" %(matrix[0,0],matrix[0,1],matrix[0,2]), file=f)
        print("%20.16f %20.16f %20.16f" %(matrix[1,0],matrix[1,1],matrix[1,2]), file=f)
        print("%20.16f %20.16f %20.16f" %(matrix[2,0],matrix[2,1],matrix[2,2]), file=f)
        element_count = {}
        for symbol in atoms.symbols: element_count[symbol] = element_count.get(symbol, 0) + 1 
        print(' '.join([str(item) for item in element_count.keys()]), file=f)
        print(' '.join([str(item) for item in element_count.values()]), file=f)
        element_count = {}
        if opt=='C':
            print('Cartesian', file=f)
            for atom in atoms:
                symbol=atom.symbol
                element_count[symbol] = element_count.get(symbol, 0) + 1
                xc, yc, zc = atom.position
                print("%20.16f %20.16f %20.16f   !%s%d" %(xc,yc,zc,symbol,element_count[symbol]), file=f)
        if opt=='D':
            print('Direct', file=f)
            mi=np.linalg.inv(matrix)
            for atom in atoms:
                symbol=atom.symbol
                element_count[symbol] += 1
                cart_coords=np.array(atom.position)
                ##CHECAR PARA SOLIDS. LA T.
                vd=np.dot(mi.T, cart_coords)
                print("%20.16f %20.16f %20.16f   !%s%d" %(vd[0], vd[1], vd[2],symbol,element_count[symbol]), file=f)
    f.close()
    #print("Writing %s" %(file))
    ##if in_log==0: print("Writing %s" %(file))
#------------------------------------------------------------------------------------------
def conventional(atoms, tol=0.01):
    lattice_vectors=np.copy(atoms.cell)
    xdmin,xdmax=0.0,1.0
    ydmin,ydmax=0.0,1.0
    zdmin,zdmax=0.0,1.0
    direct_coords = atoms.get_scaled_positions()
    atom_symbols = atoms.get_chemical_symbols()
    symbols, positions=[], []
    for i, icoord in enumerate(direct_coords):
        xd, yd, zd=icoord
        for x in [-1,0,1]:
            for y in [-1,0,1]:
                for z in [-1,0,1]:
                    xd2=xd+x
                    yd2=yd+y
                    zd2=zd+z
                    cond1=(xd2 >= xdmin-tol) and (xd2 <= xdmax+tol)
                    cond2=(yd2 >= ydmin-tol) and (yd2 <= ydmax+tol)
                    cond3=(zd2 >= zdmin-tol) and (zd2 <= zdmax+tol) 
                    if cond1 and cond2 and cond3:
                        symbols.append(atom_symbols[i])
                        vectord=np.array([xd2, yd2, zd2])
                        cart_coords=np.dot(vectord, lattice_vectors)
                        positions.append(cart_coords)
    poscarout = Atoms(symbols=symbols, positions=positions, cell=lattice_vectors, pbc=True)
    #poscarout.info['i']=atoms.info['i']
    #poscarout.info['e']=atoms.info['e']
    return poscarout
#------------------------------------------------------------------------------------------



import os.path
from ase.data import atomic_masses, atomic_numbers
#------------------------------------------------------------------------------------------
def sort_composition(composition):
    s=[[x[0],x[1], atomic_masses[atomic_numbers[x[0]]]] for x in composition]
    t = sorted(s, key=lambda x: float(x[2]), reverse=True)
    newcomposition=([[x[0],x[1]] for x in t])
    return newcomposition
#------------------------------------------------------------------------------------------
def clustername(composition):
    newcomposition=sort_composition(composition)
    chainname=' '.join([item[0]+str(item[1]) for item in newcomposition])
    return chainname
#------------------------------------------------------------------------------------------
def get_inatoms(composition):
    inatoms=[]
    newcomposition=sort_composition(composition)                                
    for xxx in newcomposition:
        for jj in range(xxx[1]):
            inatoms.append(xxx[0])
    return inatoms
#------------------------------------------------------------------------------------------
def get_elements(composition):
    newcomposition=sort_composition(composition)
    atoms=([x[0] for x in newcomposition])
    return atoms
#------------------------------------------------------------------------------------------
def read_var_composition(bilatu_file, id, index):
    bilfile=open(bilatu_file,"r")
    chainchar='---'+id.upper()+'---'
    printer=0
    data_block=[]
    for line in bilfile:
         lin = line.lstrip()
         if lin.startswith(chainchar): printer=1+printer
         if printer == 1 and not lin.startswith(chainchar):
             data_block.append(line.split())
    bilfile.close()
    nat, nvc, varcomp=len(data_block), len(data_block[0]), []
    for jj  in range(1,nvc):
        icomp=[]
        for ii in range(nat):
            atom=data_block[ii][0]
            nats=int(data_block[ii][jj])
            icomp.append([atom,nats])
        varcomp.append(icomp)
    class Composition:
        def __init__(self, xcomposition):
            self.comp    = sort_composition(xcomposition)
            self.atoms   = get_inatoms(self.comp)
            self.name    = clustername(self.comp)
            self.elements= get_elements(self.comp)
    classcomp=Composition(varcomp[index])
    return classcomp
#------------------------------------------------------------------------------------------
def read_block_of_bil(bilatu_file, id):
    bilfile=open(bilatu_file,"r")
    chainchar='---'+id.upper()+'---'
    printer=0
    data_block=[]
    for line in bilfile:
         lin = line.lstrip()
         if lin.startswith(chainchar): printer=1+printer
         if printer == 1 and not lin.startswith(chainchar): data_block.append(line)
    bilfile.close()
    return data_block 
#------------------------------------------------------------------------------------------
def get_a_float(bilatu_file, strchain, defaultvalue):
    finalvalue=float(defaultvalue)
    if os.path.isfile(bilatu_file):
        bilfile=open(bilatu_file,"r")
        for line in bilfile:
            line=line.strip(' \t\n\r')
            if len(line.strip()) != 0 :
                li = line.lstrip()
                if not li.startswith("#"):
                    readline=line.split()
                    if len(readline) == 2:
                        data0=readline[0].strip('\t\n\r') 
                        data1=readline[1].strip('\t\n\r')
                        if data0.lower() == str(strchain): finalvalue=float(data1)
        bilfile.close()
    return finalvalue
#------------------------------------------------------------------------------------------
def get_a_int(bilatu_file, strchain, defaultvalue):
    finalvalue=int(defaultvalue)
    if os.path.isfile(bilatu_file):
        bilfile=open(bilatu_file,"r")
        for line in bilfile:
            line=line.strip(' \t\n\r')
            if len(line.strip()) != 0 :
                li = line.lstrip()
                if not li.startswith("#"):
                    readline=line.split()
                    if len(readline) == 2:
                        data0=readline[0].strip('\t\n\r') 
                        data1=readline[1].strip('\t\n\r')
                        if data0.lower() == str(strchain): finalvalue=int(data1)
        bilfile.close()
    return finalvalue
#------------------------------------------------------------------------------------------
def get_a_str(bilatu_file, strchain, defaultvalue):
    finalvalue=str(defaultvalue)
    if os.path.isfile(bilatu_file):
        bilfile=open(bilatu_file,"r")
        for line in bilfile:
            line=line.strip(' \t\n\r')
            if len(line.strip()) != 0 :
                li = line.lstrip()
                if not li.startswith("#"):
                    readline=line.split()
                    if len(readline) == 2:
                        data0=readline[0].strip('\t\n\r') 
                        data1=readline[1].strip('\t\n\r')
                        if data0.lower() == str(strchain): finalvalue=str(data1)
        bilfile.close()
    return finalvalue
#------------------------------------------------------------------------------------------
def get_float_list(bilatu_file, strchain, defaultvalue):
    finalvalue=[float(item) for item in defaultvalue]
    if os.path.isfile(bilatu_file):
        bilfile=open(bilatu_file,"r")
        for line in bilfile:
            line=line.strip(' \t\n\r')
            if len(line.strip()) != 0 :
                li = line.lstrip()
                if not li.startswith("#"):
                    readline=line.split()
                    data0=readline[0].strip('\t\n\r') 
                    if data0.lower() == str(strchain):
                        del readline[0]
                        data1=[float(item) for item in readline]
                        finalvalue=data1
        bilfile.close()
    return finalvalue
#------------------------------------------------------------------------------------------
def get_int_list(bilatu_file, strchain, defaultvalue):
    finalvalue=[int(item) for item in defaultvalue]
    if os.path.isfile(bilatu_file):
        bilfile=open(bilatu_file,"r")
        for line in bilfile:
            line=line.strip(' \t\n\r')
            if len(line.strip()) != 0 :
                li = line.lstrip()
                if not li.startswith("#"):
                    readline=line.split()
                    data0=readline[0].strip('\t\n\r') 
                    if data0.lower() == str(strchain):
                        del readline[0]
                        data1=[int(item) for item in readline]
                        finalvalue=data1
        bilfile.close()
    return finalvalue
#------------------------------------------------------------------------------------------
def get_str_list(bilatu_file, strchain, defaultvalue):
    finalvalue=[str(item) for item in defaultvalue]
    if os.path.isfile(bilatu_file):
        bilfile=open(bilatu_file,"r")
        for line in bilfile:
            line=line.strip(' \t\n\r')
            if len(line.strip()) != 0 :
                li = line.lstrip()
                if not li.startswith("#"):
                    readline=line.split()
                    data0=readline[0].strip('\t\n\r') 
                    if data0.lower() == str(strchain):
                        del readline[0]
                        data1=[str(item) for item in readline]
                        finalvalue=data1
        bilfile.close()
    return finalvalue
#------------------------------------------------------------------------------------------



import os
from ase import Atom, Atoms
from ase.data import chemical_symbols
#------------------------------------------------------------------------------------------
def syva(moleculeglomos):
    natoms=len(moleculeglomos)
    namein=moleculeglomos.info['i']
    energy=moleculeglomos.info['e']
    fh=open('syvamolinput',"w")
    print(namein, file=fh)
    print(natoms, file=fh)
    for iatom in moleculeglomos:
        ani = iatom.number
        xc, yc, zc = iatom.position
        print("%-2d %16.9f %16.9f %16.9f" %(ani, xc, yc, zc), file=fh)
    fh.close()
    os.system("syva all syvamolinput > syvamolout")
    syvafile=open('syvamolout','r')
    moleculeout=[]
    for line in syvafile:
        if "ERROR" in line:
            moltmp=moleculeglomos.copy()
            moltmp.info['c'] = 'X'
            return moltmp
        if "Optimized" in line:
            hls=line.split()
            pg=str(hls[1])
            moltmp = Atoms()
            moltmp.info['e'] = energy
            moltmp.info['i'] = namein
            moltmp.info['c'] = pg
            line=syvafile.readline()
            for ii in range(natoms):
                line=syvafile.readline()
                ls=line.split()
                ss=chemical_symbols[int(ls[0])]
                xc,yc,zc = float(ls[1]), float(ls[2]), float(ls[3])
                ai=Atom(symbol=ss, position=(xc, yc, zc))
                moltmp.append(ai)
            moleculeout.extend([moltmp])
    syvafile.close()
    os.system("rm -f syvamolinput syvamolout")
    return moleculeout[-1]
#------------------------------------------------------------------------------------------
def sym_syva(moleculelist):
    moleculeout=[]
    for imol in moleculelist:
        mol1=syva(imol)
        pg=mol1.info['c']
        print('%-11s run syva:%-3s' %(mol1.info['i'], pg))
        moleculeout.extend([mol1])
    return moleculeout
#------------------------------------------------------------------------------------------


import time
import numpy as np
from aegon.libdescriptors import descriptors_sel, descriptors_comb
from aegon.libclustering import clustering_agg, clustering_hdbscan, clustering_kmeans
from aegon.libutils import sort_by_energy
#----------------Energias promedio---------------------------------------------
def selectbyclustering_uni_prom(mol_in, selection_count = [10, 8, 6, 4, 2], descriptor_type="MBTR_dis_2"):
    start = time.time()
    nn=len (mol_in)
    ni=sum(selection_count)
    if nn <= ni:
        return mol_in
    n_count= len (selection_count)
    des = descriptors_sel(mol_in, descriptor_type)    
    clus = clustering_agg(mol_in, des, n_count)    
    ord_clus = ordenar_diccionario_por_energia_promedio(clus)
    lista_est = min_de_cada_clus_con_select_count(ord_clus, selection_count)
    mol_out = sort_by_energy(lista_est, 1)
    nf=len(mol_out)
    end = time.time()
    print('Clustering selection (serial) at %5.2f s [%d -> %d]' %(end - start, nn, nf))
    return mol_out

def selectbyclustering_des_comb_prom(mol_in, selection_count = [10, 8, 6, 4, 2], descriptor_type_1 = "MBTR_dis_2", descriptor_type_2 = "MBTR_arg_2"):
    start = time.time()
    nn=len (mol_in)
    ni=sum(selection_count)
    if nn <= ni:
        return mol_in
    n_count= len (selection_count)
    des = descriptors_comb(mol_in, descriptor_type_1, descriptor_type_2)
    clus = clustering_agg(mol_in, des, n_count)
    ord_clus = ordenar_diccionario_por_energia_promedio(clus)
    lista_est = min_de_cada_clus_con_select_count(ord_clus, selection_count)
    mol_out = sort_by_energy(lista_est, 1)
    nf=len(mol_out)
    end = time.time()
    print('Clustering selection (serial) at %5.2f s [%d -> %d]' %(end - start, nn, nf))
    return mol_out
#---------------------------------por enegia del individuo de menor enegia--------------
def selectbyclustering_uni(mol_in, selection_count = [10, 8, 6, 4, 2], descriptor_type="MBTR_dis_2"):
    start = time.time()
    nn=len (mol_in)
    ni=sum(selection_count)
    if nn <= ni:
        return mol_in
    n_count= len (selection_count)
    des = descriptors_sel(mol_in, descriptor_type)    
    clus = clustering_agg(mol_in, des, n_count)    
    ord_clus = ordenar_diccionario_por_minima_energia(clus)
    lista_est = min_de_cada_clus_con_select_count(ord_clus, selection_count)
    mol_out = sort_by_energy(lista_est, 1)
    nf=len(mol_out)
    end = time.time()
    print('Clustering selection (serial) at %5.2f s [%d -> %d]' %(end - start, nn, nf))
    return mol_out

def selectbyclustering_des_comb(mol_in, selection_count = [10, 8, 6, 4, 2], descriptor_type_1 = "MBTR_dis", descriptor_type_2 = "MBTR_ang"):
    start = time.time()
    nn=len (mol_in)
    ni=sum(selection_count)
    if nn <= ni:
        return mol_in
    n_count= len (selection_count)
    des = descriptors_comb(mol_in, descriptor_type_1, descriptor_type_2)
    clus = clustering_agg(mol_in, des, n_count)
    ord_clus = ordenar_diccionario_por_minima_energia(clus)
    lista_est = min_de_cada_clus_con_select_count(ord_clus, selection_count)
    mol_out = sort_by_energy(lista_est, 1)
    nf=len(mol_out)
    end = time.time()
    print('Clustering selection (serial) at %5.2f s [%d -> %d]' %(end - start, nn, nf))
    return mol_out
#-------------------------HDBSCAN-------------------------------------------------------
def selectbyclustering_uni_hdbscan(mol_in, n_seed = 45, descriptor_type = "MBTR_dis_2"):
    start = time.time()
    nn=len (mol_in)
    if nn <= 100:
        return mol_in
    des = descriptors_sel(mol_in, descriptor_type)
    clus, n_clus =  clustering_hdbscan(mol_in, des)
    lista_est, g_fit, selection_counts = select_individuals(clus, n_seed)
    mol_out = sort_by_energy(lista_est, 1)
    nf=len(mol_out)
    end = time.time()
    print('Clustering selection (serial) at %5.2f s [%d -> %d]' %(end - start, nn, nf))
    return mol_out

def selectbyclustering_comb_hdbscan(mol_in, n_seed  = 45,  descriptor_type_1 = "MBTR_dis_2", descriptor_type_2 = "MBTR_arg_2"):
    start = time.time()
    nn=len (mol_in)
    if nn <= 100:
        return mol_in
    des = descriptors_comb(mol_in, descriptor_type_1, descriptor_type_2)
    clus, n_clus =  clustering_hdbscan(mol_in, des)
    lista_est, g_fit, selection_counts = select_individuals(clus, n_seed)
    mol_out = sort_by_energy(lista_est, 1)
    nf=len(mol_out)
    end = time.time()
    print('Clustering selection (serial) at %5.2f s [%d -> %d]' %(end - start, nn, nf))
    return mol_out
#------------------------Manejo de diccionarios--------------------------------
def ordenar_diccionario_por_energia_promedio(diccionario):
    lista_claves_energias = []
    for clave, ilistmol in diccionario.items():
        if isinstance(ilistmol, list):
            energia_total = sum(mol.info['e'] for mol in ilistmol)
            energia_media = energia_total / len(ilistmol) if ilistmol else float('inf')
            lista_claves_energias.append((clave, energia_media))
        else:
            print(f"Advertencia: {clave} no es una lista.")
    lista_claves_energias.sort(key=lambda x: x[1])
    diccionario_ordenado = {clave: diccionario[clave] for clave, _ in lista_claves_energias}
    return diccionario_ordenado

def ordenar_diccionario_por_minima_energia(diccionario):
    lista_claves_energias = []
    for clave, ilistmol in diccionario.items():
        if isinstance(ilistmol, list):
            energias = [mol.info['e'] for mol in ilistmol if 'e' in mol.info]
            energia_minima = min(energias) if energias else float('inf')
            lista_claves_energias.append((clave, energia_minima))
        else:
            print(f"Advertencia: {clave} no es una lista y será ignorado.")

    lista_claves_energias.sort(key=lambda x: x[1])
    diccionario_ordenado = {clave: diccionario[clave] for clave, _ in lista_claves_energias}
    return diccionario_ordenado
#------------------------ selection counts-------------------------------------
def min_de_cada_clus (diccionario):
    nueva_lista= []
    recorrer_dic_and_take_min(diccionario, nueva_lista)
    return nueva_lista

def min_de_cada_clus_con_select_count(diccionario,selection_count):
    nueva_lista= []
    recorrer_dic_and_take_min_con_select_count(diccionario, nueva_lista, selection_count)
    return nueva_lista

def recorrer_dic_and_take_min (diccionario, nueva_lista):
    for clave,valor in diccionario.items():
        if isinstance(valor, dict):
            recorrer_dic_and_take_min(valor, nueva_lista)
        elif isinstance( valor, list):
            list_ord = sort_by_energy(valor,1)
            nueva_lista.append(list_ord[0])

def recorrer_dic_and_take_min_con_select_count(diccionario, nueva_lista, selection_count):
    indice = 0  # Empezamos con el primer valor de la lista de cantidades
    for clave, valor in diccionario.items():
        if isinstance(valor, dict):
            recorrer_dic_and_take_min(valor, nueva_lista, selection_count)
        elif isinstance(valor, list):
            list_ord = sort_by_energy(valor, 1)
            count = selection_count[indice] if indice < len(selection_count) else 1
            nueva_lista.extend(list_ord[:count])
            indice += 1

def clust_ite(diccionario, descriptor_func, selection_count):
    nue_dic = {}
    listas = list(diccionario.values())
    for idx, (lista, n_clusters) in enumerate(zip(listas, selection_count)):
        #print(f"Procesando lista {idx + 1} con n_clusters={n_clusters}...")
        if len(lista) > 1 and len(lista) >= n_clusters:
           des = descriptor_func(lista)
           clus = clustering_agg(lista, des, n_clusters)
           nue_dic[idx] = clus
        else:
            nue_dic[idx] = {0: lista}
    return nue_dic
#----------------------------selection HDBSCAN---------------------------------
def compute_group_fitness(diccionario):
    group_avg_energies = {}
    for key, molecules in diccionario.items():
        if molecules:
            energies = [mol.info['e'] for mol in molecules]
            group_avg_energies[key] = sum(energies) / len(energies)
        else:
            group_avg_energies[key] = float('inf')

    E_min = min(group_avg_energies.values())
    E_max = max(group_avg_energies.values())

    group_fitness = {}
    for key, avgE in group_avg_energies.items():
        if E_max == E_min:
            fitness = 1.0
        else:
            normalized = (avgE - E_min) / (E_max - E_min)
            fitness = 0.5 * (1.0 - np.tanh((2.0 * normalized) - 1.0))
        group_fitness[key] = fitness
    return group_fitness

def selection_counts_from_fitness(group_fitness, total_individuals):
    total_fitness = sum(group_fitness.values())
    raw_counts = {key: (total_individuals * (fit / total_fitness)) if total_fitness != 0 else 0
                  for key, fit in group_fitness.items()}
    
    selection_counts = {key: int(np.floor(count)) for key, count in raw_counts.items()}
    allocated = sum(selection_counts.values())
    remainder = total_individuals - allocated

    sorted_keys = sorted(raw_counts, key=lambda k: raw_counts[k] - selection_counts[k], reverse=True)
    for key in sorted_keys:
        if remainder <= 0:
            break
        selection_counts[key] += 1
        remainder -= 1

    return selection_counts

def select_individuals(diccionario, total_individuals):
    group_fitness = compute_group_fitness(diccionario)
    selection_counts = selection_counts_from_fitness(group_fitness, total_individuals)

    nueva_lista = []
    for key, molecules in diccionario.items():
        sorted_molecules = sort_by_energy(molecules, 1)
        count = selection_counts.get(key, 1)
        nueva_lista.extend(sorted_molecules[:count])

    return nueva_lista, group_fitness, selection_counts



    



import os
import time
import numpy as np
from ase import Atom
from multiprocessing import Process
from aegon.libpymatgen import inequivalent_finder
from aegon.libutils import rename, adjacency_matrix, readxyzs, writexyzs, prepare_folders, split_poscarlist
#from aegon.libdiscusr import molin_sim_molref
#------------------------------------------------------------------------------------------
def neighbor_finder(atoms):
    '''
    This function builds a dictionary that contains the neighbors of all atoms
    
    in: adj_mtx (list); a list of list elements that represent the adjacency matrix 
        of the molecule as a graph.
    out: dict_neig (dict); a dictionary containing as keys the position of the atom 
        in the molecule.atoms list, and as values a list of all its neighbors.
        ({0:[1,3,7,11], 1:[0,7],...})
    '''
    dict_neig = {}
    conta = 0
    adj_mtx = adjacency_matrix(atoms)
    for i in adj_mtx:
        neig = []
        contb = 0
        for j in i:
            if j == 1:
                neig.append(contb)
            contb = contb + 1
        dict_neig[conta] = neig
        conta = conta + 1
    return dict_neig
#------------------------------------------------------------------------------------------
def triangle_finder(neighbors_dict,atom):
    '''
    This function finds the triangles formed by the binded atoms in a molecule
    in: neighbors_dict (Dict), the keys are position of the atom in the Molecule.atoms list, 
        and the values are a list of all their neighbors
        atom (int), the atom whose triangles are to be found
    out: triangles (list), a list of all 3 atoms that form each triangle 
    '''
    triangles = []
    # For this atom find its neighbors
    neiga = neighbors_dict[atom]
    for n in neiga:
        # For every neighbor find its own neighbors
        neigb = neighbors_dict[n]
        for n2 in neigb:
            # Check if this neighbor is in common with the original and if so save it
            if n2 in neiga:
                auxlist = [atom,n,n2]
                auxlist.sort()
                if auxlist in triangles:
                    continue
                triangles.append(auxlist)
    return triangles
#------------------------------------------------------------------------------------------
def focused_expansion(molin, vector):
    molout = molin.copy()
    rcoords=[]
    for a in molout.positions:
        r = np.array(a) - vector
        delta = np.linalg.norm(r)
        rnorm = r/delta
        edel = np.exp(-delta)*1.2
        a = np.array(a) + rnorm*edel
        rcoords.append(a)
    molout.positions =rcoords
    return molout
#------------------------------------------------------------------------------------------
def add_all_interstitial_atoms(original_molecule, neighbors_dict, specie):
    visited_t = []
    xxx_mol = original_molecule.copy()
    atom_list = range(len(original_molecule))
    for a in atom_list:
        # find the triangles composed by each atom in the list
        triangles = triangle_finder(neighbors_dict,a)
        for t in triangles:
            if t in visited_t:
                continue
            add_mol = original_molecule.copy()
            # get all the three vectors and its middle point
            atm1 = add_mol[t[0]].position
            atm2 = add_mol[t[1]].position
            atm3 = add_mol[t[2]].position
            v1 = np.array(atm1)
            v2 = np.array(atm2)
            v3 = np.array(atm3)
            mid_vect = (v1 + v2 + v3) / 3
            add_atom = Atom(symbol=specie, position=(mid_vect[0],mid_vect[1],mid_vect[2]))
            xxx_mol.append(add_atom)
            visited_t.append(t)
    return xxx_mol
#------------------------------------------------------------------------------------------
def add_ineq_interstitial_atoms_with_expansion(original_molecule, atom_list):
    molist_out = []
    for add_atom in atom_list:
        mid_vect = np.array(add_atom.position)
        add_mol = original_molecule.copy()
        exp_mol = focused_expansion(add_mol, mid_vect)
        exp_mol.append(add_atom)
        molist_out.append(exp_mol)
    return molist_out
#------------------------------------------------------------------------------------------
def growpal(molist_in, specie):
    '''
    This function masters all the above to add a new atom to each of the inequivalent atoms that belong
    to the outter layer of the molecule. It uses two types of addition: a.- Directly above of the atoms and 
    b.- in between two adjacent neighbors.
    in: molist_in (list[Molecule]); all the molecules that will receive an extra atom
        species (str); the species of the atom to be additioned
    out: molist_out (list[Molecule]);  a list that contains all the molecules with the extra atom 
    '''
    molist_out = []
    for imol in molist_in:
        org_nnn = len(imol)
        org_mol = imol.copy()
        org_mol.info['e'] = 0.0
        org_mol.translate(-org_mol.get_center_of_mass())
        ##FIND A DICT WITH ALL THE NEIGHBORS
        all_neighbors = neighbor_finder(org_mol)
        ##ADD specie IN ALL POSSIBLE TRIANGLES. WITHOUT ANY EXPANSION
        xxx_mol = add_all_interstitial_atoms(org_mol, all_neighbors, specie)
        inequivalent = inequivalent_finder(xxx_mol)
        ##BUILD THE LIST OF ADDED-INEQUIVALENT atom OBJECTS
        elegible_atoms=[xxx_mol[i] for i in inequivalent if i >= org_nnn]
        ##STRUCTURE WITH EXPANSIONS FOR THE ADDED-INEQUIVALENT ATOMS
        mod_mol=add_ineq_interstitial_atoms_with_expansion(org_mol, elegible_atoms)
        mod_mol=rename(mod_mol, imol.info['i'], 4)
        molist_out.extend(mod_mol)
    return molist_out
#------------------------------------------------------------------------------------------
def growpal_serial(poscarlist, specie):
    start = time.time()
    atoms1=growpal(poscarlist, specie)
    end = time.time()
    print('GrowPal generation at %5.2f s' %(end - start))
    return atoms1
#------------------------------------------------------------------------------------------
def make_growpal(ifolder, specie):
    atoms0=readxyzs(ifolder+'/'+ifolder+'.xyz')
    atoms1=growpal(atoms0, specie)
    writexyzs(atoms1,ifolder+'/'+ifolder+'_growpal.xyz')
    return 0
#------------------------------------------------------------------------------------------
def growpal_parallel(poscarlist, specie, nproc, base_name, tol=0.95, n_features=16):
    start = time.time()
    if not isinstance(poscarlist, list): poscarlist = [poscarlist]
    folderlist=prepare_folders(poscarlist, nproc, base_name)
    poscar_split_list=split_poscarlist(poscarlist, nproc)
    procs = []
    for ifolder, iposcars in zip(folderlist, poscar_split_list):
        writexyzs(iposcars, ifolder+'/'+ifolder+'.xyz')
        proc = Process(target=make_growpal, args=(ifolder,specie,))
        procs.append(proc)
        proc.start()
    for proc in procs:
        proc.join()
    moleculeout=[]
    for ifolder in folderlist:
        molx=readxyzs(ifolder+'/'+ifolder+'_growpal.xyz')
        #molx=molin_sim_molref(molx,moleculeout, tol, n_features)
        moleculeout=moleculeout+molx
    end = time.time()
    print('GrowPal (parallel) at %5.2f s [%d]' %(end - start, len(moleculeout)))
    os.system('rm -rf %sproc[0-9][0-9]' %(base_name))
    return moleculeout
#------------------------------------------------------------------------------------------
#from aegon.libsyva  import sym_syva
#from aegon.libutils import readxyzs, writexyzs
#mol1=readxyzs('b16.xyz')
#mol1=sym_syva(mol1)
#mol2=make_many_rand(mol1, 'Be')
#writexyzs(mol2,'out.xyz')



from sklearn.cluster import KMeans, AgglomerativeClustering, HDBSCAN
#----------------------------------------------------------------------------------------------------------

def clustering_agg(lista, descriptors, n_clusters):

    if len(descriptors) < n_clusters:
        print("La cantidad de descriptores es menor que el número de clusters. Devolviendo la lista original.")
        return {0: lista}
    else:
        model = AgglomerativeClustering(n_clusters=n_clusters)
        labels = model.fit_predict(descriptors)
        
    clusters = {i: [] for i in range(n_clusters)}
    for idx, label in enumerate(labels):
        clusters[label].append(lista[idx])
    
    return clusters
#-----------------------------------------------------------------------------------------------------------

def clustering_kmeans (lista, descriptors, n_clusters):
    
    if len(descriptors) < n_clusters:
        print("La cantidad de descriptores es menor que el número de clusters. Devolviendo la lista original.")
        return {0: lista}
    else:
        model = KMeans(n_clusters=n_clusters, init='k-means++', max_iter=3000, tol=1e-7)
        model.fit(descriptors)
        labels = model.labels_
        
    clusters = {i: [] for i in range(n_clusters)}
    for idx, label in enumerate(labels):
        clusters[label].append(lista[idx])

    return clusters
#--------------------------------------------------------------------------------------------------------------
    
def clustering_hdbscan(lista, descriptores):
    if len(descriptores) < 100:
        print("La cantidad de descriptores es menor a 100. Devolviendo la lista original.")
        return {0: lista}
    
    model = HDBSCAN(min_cluster_size = 5, cluster_selection_epsilon = 0.2, cluster_selection_method = "eom")
    #model = DBSCAN()
    model.fit(descriptores)

    labels = model.labels_
    n_clus = len(set(labels))

    clusters = {label: [] for label in set(labels)}  

    for idx, label in enumerate(labels):
        clusters[label].append(lista[idx])

    return clusters, n_clus



import numpy as np
from rdkit import Chem
from ase.data import covalent_radii
#------------------------------------------------------------------------------------------
def adjmatrixfromsdfconectivity(rdkitmol):
    natoms=rdkitmol.GetNumAtoms()
    matrixcd=np.zeros(shape=(natoms,natoms),dtype=np.int64)
    for bond in rdkitmol.GetBonds():
       i=bond.GetBeginAtomIdx()
       j=bond.GetEndAtomIdx()
       o=bond.GetBondTypeAsDouble()
       if o > 0.0:
           matrixcd[i][j] = int(1)
           matrixcd[j][i] = int(1)
    return matrixcd
#------------------------------------------------------------------------------------------
def adjmatrixfromsdfpositions(rdkitmol, factor=1.2):
    natoms=rdkitmol.GetNumAtoms()
    conf = rdkitmol.GetConformer()
    matrixc=np.zeros(shape=(natoms,natoms),dtype=np.int64)
    for i in range(natoms):
        si=rdkitmol.GetAtomWithIdx(i).GetSymbol()
        ri=get_covalent_radius(si)
        vi=conf.GetAtomPosition(i)
        xi,yi,zi=float(vi.x), float(vi.y), float(vi.z)
        for j in range(i+1,natoms):
            sj=rdkitmol.GetAtomWithIdx(j).GetSymbol()
            rj=get_covalent_radius(sj)
            vj=conf.GetAtomPosition(j)
            xj,yj,zj=float(vj.x), float(vj.y), float(vj.z)
            xx,yy,zz=xj-xi, yj-yi, zj-zi
            distance = np.sqrt(xx*xx+yy*yy+zz*zz)
            dist = distance/(ri + rj)
            if ( dist <= factor ):
                matrixc[i][j] = int(1)
                matrixc[j][i] = int(1)
    return matrixc
#------------------------------------------------------------------------------------------
def rdkitmol2asemol(rdkitmol):
    symbols=[atom.GetSymbol() for atom in rdkitmol.GetAtoms()] 
    positions=rdkitmol.GetConformer().GetPositions()
    asemol=Atoms(symbols=symbols, positions=positions)
    return asemol
#------------------------------------------------------------------------------------------
def asemol2rdkitmol(molase, factor=1.2):
    natoms=molase.get_global_number_of_atoms()
    symbols=molase.get_chemical_symbols()
    r=[covalent_radii[k] for k in molase.numbers]
    positions=molase.get_positions()
    molrdkit = Chem.RWMol()
    conformer = Chem.Conformer(natoms)
    bond_info=[]
    for i in range(natoms):
        atom = Chem.Atom(symbols[i])
        molrdkit.AddAtom(atom)
        conformer.SetAtomPosition(i, positions[i][:])
        for j in range(i+1,natoms):
            dij=molase.get_distances(i,j)/(r[i] + r[j])
            if ( dij <= factor ):
                bond_info.append((i, j, Chem.rdchem.BondType.SINGLE))
    molrdkit.AddConformer(conformer)
    for bond in bond_info:
        atom1_idx, atom2_idx, bond_type = bond
        molrdkit.AddBond(atom1_idx, atom2_idx, bond_type)
    return molrdkit
#------------------------------------------------------------------------------------------


import os
import numpy as np 
from ase import Atom, Atoms
#------------------------------------------------------------------------------------------
hartree2eV = 27.211386245981 #NIST
bohr2angstrom=0.529177210544 #NIST
eVtokcalpermol=23.060548012069496
hartree2kcalmol=627.5094738898777
#------------------------------------------------------------------------------------------
def get_termination_orca(pathfilename):
    if os.path.isfile(pathfilename):
        normal=0
        file=open(pathfilename,'r')
        for line in file:
            if "ORCA TERMINATED NORMALLY" in line: normal=normal+1
        file.close()
        return normal
    else:
        return False
#------------------------------------------------------------------------------------------
def get_geometry_orca(pathfilename):
    nt=get_termination_orca(pathfilename)
    if nt==False: return False
    filename=os.path.basename(pathfilename)
    namein=filename.split('.')[0]
    trajectory = get_traj_orca(pathfilename,False)
    enemin = float('inf')
    for i, atoms in enumerate(trajectory):
        energy=atoms.info.get('e', 0.0)
        if energy < enemin:
            enemin=energy
            poscar=atoms.copy()
            poscar.info['c']=nt
            poscar.info['e']=energy
            poscar.info['i']=namein
    return (poscar)
#------------------------------------------------------------------------------------------
def get_traj_orca(pathfilename, force=False):
    filename = os.path.basename(pathfilename)
    namein=filename.split('.')[0]
    start, end, ene, start_2, end_2 = [], [], [], [], []
    openold = open(pathfilename,"r")
    rline = openold.readlines()
    for i in range(len(rline)):
        if "CARTESIAN COORDINATES (ANGSTROEM)" in rline[i]:
            start.append(i+2)
            for j in range(i + 2, len(rline)):
                if rline[j].strip() == "":
                    end.append(j - 1)
                    break
        if "TOTAL SCF ENERGY" in rline[i]:
            eneline = rline[i+3].split()
            ene.append(eneline[5]) #ENERGY IN ELECTROVOLT

        if "CARTESIAN GRADIENT" in rline[i] and force:
            start_2.append(i+3)
            for j in range(i + 3, len(rline)):
                if rline[j].strip() == "":
                    end_2.append(j - 1)
                    break
    moleculeout=[]
    for i,iStart in enumerate(start[:-1]):
        singlemol = Atoms()
        singlemol.info['e'] = float(ene[i]) #ENERGY IN ELECTROVOLT
        singlemol.info['i'] = namein+'_'+str(i+1).zfill(3)
        for line in rline[start[i] : end[i]+1]:
            words = line.split()
            ss = str(words[0])
            xc,yc,zc = float(words[1]), float(words[2]), float(words[3])
            ai=Atom(symbol=ss,position=(xc, yc, zc))
            singlemol.append(ai)
        if force:
            forces_list_by_group = []
            for line in rline[start_2[i] : end_2[i]+1]:
                words = line.split()
                fx,fy,fz = float(words[3]), float(words[4]), float(words[5])
                fx=-fx*hartree2eV/bohr2angstrom
                fy=-fy*hartree2eV/bohr2angstrom
                fz=-fz*hartree2eV/bohr2angstrom
                #IN ev/A
                forces_list_by_group.append([fx,fy,fz])    
            singlemol.arrays['forces'] = np.array(forces_list_by_group)
        moleculeout.extend([singlemol])
    openold.close()
    return (moleculeout)
#------------------------------------------------------------------------------------------
